/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.parser.api;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.log.parser.api.model.Flow;
import org.ow2.petals.log.parser.api.model.FlowId;
import org.ow2.petals.log.parser.api.model.FlowStep;

/**
 * A class used to parse and organize Petals logged data.
 * <p>
 * Since Petals 4.2, MONIT-MSG entries have been removed.
 * They are now MONIT entries created by another logger, with additional fields.
 * </p>
 * <p>
 * Usage:
 * <code><pre>
 * FlowBuilder flowBuilder = new FlowBuilder( dir1, dir2, ... );
 * Set<String> flowIds = flowBuilder.findFlowIds();
 * for( String flowId : flowIds ) {
 *      Flow flow = flowBuilder.parseFlow( flowId );
 *      // Do something with the flow
 * }
 * </pre></code>
 * </p>
 *
 * @author Pierre-Yves Gibello - EBM WebSourcing
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class FlowBuilder {

    private File[] logDirectories;
    private final Logger logger;


    /**
     * Constructor.
     */
    public FlowBuilder() {
        this.logger = Logger.getLogger( FlowBuilder.class.getName());
    }


	/**
	 * @return the logDirectories
	 */
	public File[] getLogDirectories() {
		return this.logDirectories;
	}


	/**
     * Sets the log directories.
     * @param logDirectories a list of directories containing Petals logs (not null, not empty)
     */
    public void setLogDirectories(  File... logDirectories  ) {

    	if( logDirectories == null || logDirectories.length == 0 )
            throw new IllegalArgumentException( "At least one directory must be passed as parameter." );

        this.logDirectories = logDirectories;
        for( File dir : this.logDirectories ) {
            if( dir == null )
                throw new IllegalArgumentException( "A null directory was passed in the list." );

            if( ! dir.isDirectory())
                throw new IllegalArgumentException( dir + " is not a valid directory." );
        }
    }


    /**
     * Sets the log directories.
     * @param logDirectories a list of directory paths containing Petals logs (not null, not empty)
     */
    public void setLogDirectories( String... logDirectories ) {
    	setLogDirectories( convertToFiles( logDirectories ));
    }


    /**
     * Parses a flow directory's content and builds a {@link Flow} instance.
     * @param flowId a flow ID (not null, not the empty string)
     * @return a flow, or null if it could not be built
     */
    public Flow parseFlow( String flowId ) {

        if( flowId == null || flowId.trim().length() == 0 )
            throw new IllegalArgumentException( "A flow ID cannot be null or empty." );

        // Index all the MONIT steps
        Map<String,FlowStep> flowStepIdToFlowStep = new HashMap<String,FlowStep> ();
        Map<String,String> flowStepIdToPreviousStepId = new HashMap<String,String> ();
        Collection<String> rawMonitEntries = readRawMonitEntries( flowId );

        if( this.logger.isLoggable( Level.FINE ))
            this.logger.fine( "Indexing MONIT steps for " + flowId );

        for( String rme : rawMonitEntries ) {
        	String[] values = rme.split( PetalsLogConstants.SEP );
        	if( values.length != 2 )
        		continue;

            HashMap<String,String> properties = parseStep( values[ 1 ]);

            String flowStepId = properties.get(PetalsExecutionContext.FLOW_STEP_ID_PROPERTY_NAME);
            FlowStep flowStep = flowStepIdToFlowStep.get( flowStepId );
            if( flowStep == null ) {
                flowStep = new FlowStep( flowStepId );
                flowStepIdToFlowStep.put( flowStepId, flowStep );
            }

            String traceCode = properties.get(FlowLogData.FLOW_STEP_TRACE_CODE);
            if(traceCode.startsWith("consume")) {
            	flowStep.setType(FlowStep.CONSUME_FLOW_STEP);
            }

            flowStep.setLogFile( new File( values[ 0 ]));
            String dumpPath = properties.get( PetalsLogConstants.DUMP_PATH );
            if(traceCode.endsWith("Begin")) {
            	flowStep.setBeginProperties(properties);
            	flowStep.getBeginProperties().put(PetalsLogConstants.DUMP_PATH, dumpPath);

            } else {
            	flowStep.setOutcomeProperties(properties);
            	flowStep.getOutcomeProperties().put( PetalsLogConstants.DUMP_PATH, dumpPath );
            	if(traceCode.endsWith("Failure"))
            		flowStep.setFailure(true);
            }

            String previousFlowStepId = properties
                    .get(PetalsExecutionContext.FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME);
            if( previousFlowStepId != null )
                flowStepIdToPreviousStepId.put( flowStepId, previousFlowStepId );
        }

        // Build the dependencies between the flow steps
        // This step was prepared before
        if( this.logger.isLoggable( Level.FINE ))
            this.logger.fine( "Building the flow step dependencies for " + flowId );

        List<FlowStep> rootCandidates = new ArrayList<FlowStep> ();
        for( FlowStep flowStep : flowStepIdToFlowStep.values()) {
            String previousFlowStepId = flowStepIdToPreviousStepId.get( flowStep.getId());
            FlowStep parent = null;

            if( previousFlowStepId != null
                    && ( parent = flowStepIdToFlowStep.get( previousFlowStepId )) != null ) {
                parent.addChild( flowStep );
                flowStep.setParent( parent );

            } else {
            	// No parent? It may be the root.
            	rootCandidates.add( flowStep );
            }
        }


        // If multiple root candidates, filter those with a previous step ID
        // (the real one should have none... or nobody can guess which one it is !)
        if(rootCandidates.size() > 1) {
        	List<FlowStep> realCandidates = new ArrayList<FlowStep> ();
        	for(FlowStep step : rootCandidates) {
        		String previousFlowStepId = flowStepIdToPreviousStepId.get( step.getId());
        		if(previousFlowStepId == null)
        			realCandidates.add(step);
        	}

        	rootCandidates = realCandidates;
        }

        // There should be only one root
        Flow result = null;
        if( rootCandidates.size() == 1 )
            result = new Flow( flowId, rootCandidates.get( 0 ));
        else
            this.logger.warning( "Could not find the flow root step. Flow build is given up." );

        return result;
    }


	/**
     * @return a set flow IDs contained in the log directories.
     * <p>
     * The flow IDS are sorted by date (most recent first).
     * </p>
     */
    public Set<FlowId> findFlowIds() {

        if( this.logger.isLoggable( Level.FINE ))
            this.logger.fine( "Listing flow IDs." );

        // Find all the flow information
        Map<String,Long> nameToLastModif = new HashMap<String,Long> ();
    	for( File dir : this.logDirectories ) {
    		FileFilter filterDirs = new FileFilter() {
    			@Override
                public boolean accept(File file) {
    				return file.isDirectory() && !file.getName().startsWith(".");
    			}
    		};

    		File subdirs[] = dir.listFiles(filterDirs);
    		for( File sd : subdirs ) {
    			Long lSaved = nameToLastModif.get( sd.getName());
    			Long lToInsert = sd.lastModified();
    			if( lSaved != null && lSaved > lToInsert )
    				lToInsert = lSaved;

    			nameToLastModif.put( sd.getName(), lToInsert );
    		}
    	}

    	// Now, store and order it
    	Set<FlowId> sortedFlowIds = new TreeSet<FlowId>( new Comparator<FlowId> () {
        	@Override
        	public int compare( FlowId o1, FlowId o2 ) {
        		return (int) (o1.getLastModification() - o2.getLastModification());
        	}
        });

    	for( Map.Entry<String,Long> entry : nameToLastModif.entrySet()) {
    		FlowId flowId = new FlowId();
			flowId.setName( entry.getKey());
			flowId.setLastModification( entry.getValue());
			sortedFlowIds.add( flowId );
    	}

    	return sortedFlowIds;
    }


    /**
     * Reads the raw MONIT entries in the petals.log file.
     * <p>
     * This method is made public, but does not have to be used.<br />
     * It is already invoked when using {@link #parseFlow(String)}.
     * </p>
     *
     * @param flowId a flow ID (not null, not the empty string)
     * @return a list of raw log entries, or null if the flow ID is invalid
     * <p>
     * Every raw log entry starts with the location of the log file where this entry was read.
     * The file path is separated from the log entry by {@link PetalsLogConstants#SEP}.
     * </p>
     */
    public Collection<String> readRawMonitEntries( String flowId ) {

    	if( flowId == null || flowId.trim().length() == 0 )
    	    throw new IllegalArgumentException( "A flow ID cannot be null or empty." );

    	List<String> rawMonitEntries = new ArrayList<String> ();
    	for( File dir : this.logDirectories ) {

    		File log = new File( dir, flowId + File.separator + "petals.log" );
    		if( ! log.exists())
    		    continue;

    		BufferedReader in = null;
    		try {
    		    in = new BufferedReader( new FileReader(log));
    		    String line;
    		    while(( line = in.readLine()) != null ) {
    		        if( line.contains(" MONIT "))
    		            rawMonitEntries.add( log.getAbsolutePath() + PetalsLogConstants.SEP + line );
    		    }

    		} catch( IOException e ) {
    		    this.logger.fine( "An error occurred while reading raw MONIT entries: " + e.getMessage());

    		} finally {
    		    try {
    		        if( in != null )
    		            in.close();
    		    } catch(Exception ignore) {
    		        // nothing
    		    }
    		}
    	}

    	return rawMonitEntries;
    }


    /**
     * Parses a step and extracts its properties.
     * <p>
     * This method is made public, but does not have to be used.<br />
     * It is already invoked when using {@link #parseFlow(String)}.
     * </p>
     *
     * @param rawLogEntry a raw log entry (MONIT), not null, not the empty string
     * @return the step's properties or null if the flow step is invalid
     */
    public static HashMap<String,String> parseStep( String rawLogEntry ) {

    	if( rawLogEntry == null || rawLogEntry.trim().length() == 0 )
    	    throw new IllegalArgumentException( "A raw log entry cannot be null or empty." );

    	HashMap<String,String> flowStep = new HashMap<String,String>();
    	int pos = rawLogEntry.indexOf("] :");
    	if(pos > 0) {
    		String header = rawLogEntry.substring(0, pos+1);
    		String content = rawLogEntry.substring(pos+4);

    		// Parse header (node + date + component)
    		StringTokenizer st = new StringTokenizer(header, " \t[]");
    		int tok = 1;
    		StringBuffer buf = new StringBuffer();

    		while(st.hasMoreTokens()) {
    			switch(tok++) {
    			case 1: // Node
    				flowStep.put("node", st.nextToken());
    				break;

    			case 2: // Date
    				buf.append(st.nextToken() + " ");
    				break;

    			case 3: // Time
    				buf.append(st.nextToken() + " ");
    				break;

    			case 4: // GMT+...
    				buf.append(st.nextToken());
    				flowStep.put("date", buf.toString());
    				break;

    			case 5: // MONIT (ignore)
    				st.nextToken();
    				break;

    			case 6:
    				flowStep.put("component", st.nextToken());
    				break;

    			default:
    				st.nextToken();
    				break;
    			}
    		}

    		// Parse the Petals attributes
    		st = new StringTokenizer(content, ",");
    		while( st.hasMoreTokens()) {
    			String pair = st.nextToken();
    			pos = pair.indexOf('=');

    			if(pos >= 1) {
    				String val = pair.substring(pos+1).trim();
    				if(val.startsWith("\'"))
    				    val = val.substring(1);

    				if(val.endsWith("\'"))
    				    val = val.substring(0, val.length()-1);

    				flowStep.put(pair.substring(0, pos-1).trim(), val);
    			}
    		}
    	}

    	return flowStep;
    }


    /**
     * A method used to provide a secondary but still useful constructor.
     * @param files an array of file paths (can be null, empty, can contain null values)
     * @return an array of files
     */
    private static File[] convertToFiles( String[] filePaths ) {

        File[] result = null;
        if( filePaths != null ) {
            result = new File[ filePaths.length ];
            for( int i=0; i<filePaths.length; i++ )
                result[ i ] = filePaths[ i ] == null ? null : new File( filePaths[ i ]);
        }

        return result;
    }
}
