/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.parser.api.model;

import java.io.File;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Set;

import org.ow2.petals.log.parser.api.PetalsLogConstants;

/**
 * A flow step represents a set of log entries with the MONIT or MONIT-MSG level.
 * <p>
 * Every flow step has a unique ID.<br />
 * It matches an exchange in Petals (request / response / ACK).<br />
 * It has a "begin" part and an "end" or "failure" part.
 * </p>
 * <p>
 * A flow step may have a parent and children too.<br />
 * This relation is deduced from the {@link PetalsLogConstants#PROVIDE_FLOW_PREVIOUS_STEP_ID}
 * property.
 * </p>
 *
 * @author Pierre-Yves Gibello - EBM WebSourcing
 */
public class FlowStep {

	public static final int PROVIDE_FLOW_STEP = 0;
	public static final int CONSUME_FLOW_STEP = 1;


	/**
	 * The flow step ID.
	 */
    final String id_;

    /**
     * The type of the flow step.
     * <p>
     * Default is {@link #PROVIDE_FLOW_STEP}.
     * </p>
     */
    int type_ = PROVIDE_FLOW_STEP;

    /**
     * The properties of the BEGIN.
     */
    HashMap<String,String> begin_;

    /**
     * The properties of the END or FAULT.
     */
    HashMap<String,String> outcome_;

    /**
     * Indicates whether the outcome is a FAULT or an END.
     * <p>
     * Default is <b>false</b>, ie we assume it is an END.
     * </p>
     */
    boolean failure_;

    /**
     * The children steps (Petals exchanges that were created because of the current exchange).
     */
    final LinkedList<FlowStep> children_ = new LinkedList<FlowStep> ();

    /**
     * A map indexing the children, so that search goes faster.
     */
    HashMap<String,FlowStep> childIdToChildStep_ = new HashMap<String,FlowStep>();

    /**
     * The parent of this flow step.
     */
	FlowStep parent_;

	/**
	 * The log file, useful to find the absolute path of a dumped payload.
	 */
	File logFile_;



    /**
	 * Constructor.
	 * @param id the flow step's ID
	 */
	public FlowStep( String id ) {
		this.id_ = id;
	}


	/**
     * @return the id
     */
    public String getId() {
        return this.id_;
    }


	/**
     * @return the type_
     */
    public int getType() {
        return this.type_;
    }

    /**
     * @param type the type_ to set
     */
    public void setType(int type) {
        this.type_ = type;
    }


    /**
     * @return the begin
     */
    public HashMap<String,String> getBeginProperties() {
        return this.begin_;
    }


    /**
     * @param begin the begin to set
     */
    public void setBeginProperties(HashMap<String,String> begin) {
        this.begin_ = begin;
    }


    /**
     * @return the outcome_
     */
    public HashMap<String,String> getOutcomeProperties() {
        return this.outcome_;
    }

    /**
     * @param outcome_ the outcome_ to set
     */
    public void setOutcomeProperties(HashMap<String,String> outcome) {
        this.outcome_ = outcome;
    }

    /**
     * @return the failure_
     */
    public boolean isFailure() {
        return this.failure_;
    }

    /**
     * @param failure the failure_ to set
     */
    public void setFailure(boolean failure) {
        this.failure_ = failure;
    }


    /**
     * @return the parent
     */
    public FlowStep getParent() {
        return this.parent_;
    }


    /**
     * @return the children
     */
    public LinkedList<FlowStep> getChildren() {
        return this.children_;
    }


    /**
     * @return the number of children
     */
    public int getChildCount() {
    	return this.children_.size();
    }


    /**
     * Finds a descendant step.
     * @param flowStepId the flow step ID
     * @return the flow step, or null if it was not found
     */
    public FlowStep findDescendant( String flowStepId ) {

    	FlowStep child = this.childIdToChildStep_.get(flowStepId);
    	if(child == null) {
    		for(FlowStep step : this.children_) {
    			child = step.findDescendant(flowStepId);
    			if(child != null)
    			    break;
    		}
    	}

    	return child;
    }


    /**
     * @return the IDs of the direct children steps
     */
    public Set<String> getChildrenFlowStepIds() {
    	return this.childIdToChildStep_.keySet();
    }


    /**
	 * Sets the parent
	 * @param parent
	 */
	public void setParent(FlowStep parent) {
		this.parent_ = parent;
	}


	/**
	 * Adds a child.
	 * @param child
	 */
	public void addChild( FlowStep child ) {
		this.children_.add( child );
		this.childIdToChildStep_.put(child.getId(), child);
	}


	/**
	 * @return the logFile_
	 */
	public File getLogFile() {
		return this.logFile_;
	}


	/**
	 * @param logFile_ the logFile_ to set
	 */
	public void setLogFile( File logFile_ ) {
		this.logFile_ = logFile_;
	}


	/**
	 * Prints the flow step with its children.
	 * @param buf a string buffer (not null)
	 * @param level the level (0 for the tree root)
	 * @return a string representation of the flow step and its descendants
	 */
	protected String toString( StringBuilder buf, int level ) {

	    if( level > 0 )
		    buf.append('|');

		for( int i=0; i<level; i++ )
		    buf.append( "-" );

		buf.append( " " + this.id_ + "\n" );
    	for( FlowStep c : this.children_ )
    		c.toString( buf, level + 1 );

		return buf.toString();
	}


	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
	    return this.id_;
	}
}
