/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.parser.api;

import java.io.File;
import java.util.Collection;
import java.util.Set;

import junit.framework.Assert;

import org.junit.Test;
import org.ow2.petals.log.parser.api.miscellaneous.TestUtils;
import org.ow2.petals.log.parser.api.model.Flow;
import org.ow2.petals.log.parser.api.model.FlowId;
import org.ow2.petals.log.parser.api.model.FlowStep;

/**
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class FlowBuilderTestCase {

    /**
     * Performs tests on {@link FlowBuilder} with one node logs (without dumps).
     * @throws Exception
     */
    @Test
    public void testSingleNodeWithoutDump() throws Exception {

        // Testing the number of flow IDs
        File[] rootFiles = TestUtils.findTestDirectories( "oneNodeWithoutDump" );
        FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories( rootFiles );
        Set<FlowId> flowIds = flowBuilder.findFlowIds();
        Assert.assertEquals( 1, flowIds.size());
        Assert.assertEquals( "4332dd30-a72d-11e2-8071-001c231aad1f", flowIds.iterator().next().getName());

        // Testing the number of raw log entries (easy to count with an IDE)
        Collection<String> rawLogEntries = flowBuilder.readRawMonitEntries( flowIds.iterator().next().getName());
        Assert.assertEquals( 6, rawLogEntries.size());

        // Checking assertions on the final structure
        Flow flow = flowBuilder.parseFlow( flowIds.iterator().next().getName());
        Assert.assertNotNull( flow );

        // The root
        Assert.assertNotNull( flow.getRoot());
        Assert.assertEquals( "4332dd31-a72d-11e2-8071-001c231aad1f", flow.getRoot().getId());
        Assert.assertNotNull(flow.getRoot().getBeginProperties());
        Assert.assertNotNull(flow.getRoot().getOutcomeProperties());
        Assert.assertEquals( 1, flow.getRoot().getChildren().size());

        Assert.assertEquals(flow.getRoot().getType(), FlowStep.CONSUME_FLOW_STEP);
        Assert.assertEquals( 12, flow.getRoot().getBeginProperties().size());
        Assert.assertEquals( "http://10.0.0.1:8084/petals/services/Echo", flow.getRoot().getBeginProperties().get( PetalsLogConstants.SOAP_CONSUME_REQUESTED_URL ));

        Assert.assertFalse(flow.getRoot().isFailure());
        Assert.assertEquals( 7, flow.getRoot().getOutcomeProperties().size());

        // The first (and single) child
        FlowStep firstChild =  flow.getRoot().getChildren().iterator().next();
        Assert.assertEquals( flow.getRoot(), firstChild.getParent());
        Assert.assertEquals( 0, firstChild.getChildren().size());
        Assert.assertNotNull(firstChild.getBeginProperties());
        Assert.assertNotNull(firstChild.getOutcomeProperties());
        Assert.assertFalse(firstChild.isFailure());
    }


    /**
     * Performs tests on {@link FlowBuilder} with one node logs (with dumps).
     * @throws Exception
     */
    @Test
    public void testSingleNodeWithDump() throws Exception {

        // Testing the number of flow IDs
        File[] rootFiles = TestUtils.findTestDirectories( "oneNodeWithDump" );
        FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories( rootFiles );
        Set<FlowId> flowIds = flowBuilder.findFlowIds();
        Assert.assertEquals( 1, flowIds.size());
        Assert.assertEquals( "9ed1b7c0-a72c-11e2-bb5b-001c231aad1f", flowIds.iterator().next().getName());

        // Testing the number of raw log entries (easy to count with an IDE)
        Collection<String> rawLogEntries = flowBuilder.readRawMonitEntries( flowIds.iterator().next().getName());
        Assert.assertEquals( 6, rawLogEntries.size());

        // Checking assertions on the final structure
        Flow flow = flowBuilder.parseFlow( flowIds.iterator().next().getName());
        Assert.assertNotNull( flow );

        // The root
        Assert.assertNotNull( flow.getRoot());
        Assert.assertEquals( "9ed1b7c1-a72c-11e2-bb5b-001c231aad1f", flow.getRoot().getId());
        Assert.assertNotNull(flow.getRoot().getBeginProperties());
        Assert.assertNotNull(flow.getRoot().getOutcomeProperties());
        Assert.assertEquals( 1, flow.getRoot().getChildren().size());

        Assert.assertEquals(flow.getRoot().getType(), FlowStep.CONSUME_FLOW_STEP);
        Assert.assertEquals( 12, flow.getRoot().getBeginProperties().size());
        Assert.assertEquals( "http://10.0.0.1:8084/petals/services/Echo", flow.getRoot().getBeginProperties().get( PetalsLogConstants.SOAP_CONSUME_REQUESTED_URL ));

        Assert.assertTrue(flow.getRoot().isFailure());
        Assert.assertEquals( 8, flow.getRoot().getOutcomeProperties().size());

        // The first (and single) child
        FlowStep firstChild =  flow.getRoot().getChildren().iterator().next();
        Assert.assertEquals( flow.getRoot(), firstChild.getParent());
        Assert.assertNotNull(firstChild.getBeginProperties());
        Assert.assertNotNull(firstChild.getOutcomeProperties());
        Assert.assertTrue(firstChild.isFailure());

        String dumpPath = firstChild.getOutcomeProperties().get( PetalsLogConstants.DUMP_PATH );
        Assert.assertNotNull( dumpPath );
        File dumpFile = PetalsLogUtils.resolveDumpFileLocation( firstChild, false );
        Assert.assertNotNull( dumpFile );
    }


    /**
     * Test flow sorting by date (most recent first).
     * @throws Exception
     */
    @Test
    public void testFlowSorting() throws Exception {

        // Testing the number of flow IDs
        File[] rootFiles = TestUtils.findTestDirectories( "severalflows" );
        FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories( rootFiles );

        Set<FlowId> flowIds = flowBuilder.findFlowIds();
        Long lastModified = 0L;
        for( FlowId flowId : flowIds ) {
        	Assert.assertTrue( flowId.getLastModification() > lastModified );
        	lastModified = flowId.getLastModification();
        }
    }
}
